//
// $Id: HTKfile.cc,v 1.1.1.1 2005/05/06 01:15:31 xanguera Exp $
//
#include <string>
#include <vector>
#include <numeric>
#include <iostream>
#include <fstream>
#include <math.h>

#include "global.h"
#include "windowBounds.h"
#include "HTKfile.h"

using namespace std;

// Run Cepstral Mean Subtraction on the (speech-only) feature vectors.
// We create a separate mean and a variance for each feature vector.
// The means and vars are computed over a window of length 'winlen' vectors.
// Window boundary conditions are handled by the 'windowBounds' object.
void 
HTKfile::run_CMS(unsigned int winlen) {

  if (!spOnlyFlag) {	      // haven't created the spOnly[] data yet
    vector<bool> spflags(get_nObs(),true); // assume all vectors are speech
    fillSPOnly(spflags);
  }

  // Create a vector of window boundary objects, one for each feat vector.
  vector<windowBounds> winBounds;
  getWinBounds(winBounds, winlen);

  // There should be as many window boundary objs as there are frames of data
  if (winBounds.size() != get_nObs()) {
    cerr << "HTKfile::run_CMS(): num windows != num observations" << endl;
    exit(1);
  }

  // Storage for the means.
  vector<double> means(get_nComps());
  
  // Create the data structure to hold the mean-subtracted features
  vector<vector<float> > normedFeats(get_nObs());

  // Create the normalized features
  vector<float> tmpVec(get_nComps()); // tmp vector for the CMS features
  for(int i = 0; i<get_nObs(); i++) {
    getMean(winBounds[i],means); // compute mean for this window
    vector<float>& curVec = spOnly[i]; // get current feature vector
    for(int dim=0;dim<curVec.size();dim++) { 
      tmpVec[dim] = curVec[dim] - means[dim]; // do mean subtraction
    }
    // save the normalized feature vector
    normedFeats[i].assign(tmpVec.begin(),tmpVec.end());	
  }

  // Replace the unnormalized features with the normalized features
  spOnly.assign(normedFeats.begin(),normedFeats.end());

}

// Run Cepstral Mean Normalization on the (speech-only) feature vectors.
// We create a separate mean and a variance for each feature vector.
// The means and vars are computed over a window of length 'winlen' vectors.
// Window boundary conditions are handled by the 'windowBounds' object.
void 
HTKfile::run_CMN(unsigned int winlen) {

  if (!spOnlyFlag) {	      // haven't created the spOnly[] data yet
    vector<bool> spflags(get_nObs(),true); // assume all vectors are speech
    fillSPOnly(spflags);
  }

  // Create a vector of window boundary objects, one for each feat vector.
  vector<windowBounds> winBounds;
  getWinBounds(winBounds, winlen);

  // There should be as many window boundary objs as there are frames of data
  if (winBounds.size() != get_nObs()) {
    cerr << "HTKfile::run_CMS(): num windows != num observations" << endl;
    exit(1);
  }

  // Storage for the means and variances. Reused for each frame.
  vector<double> means(get_nComps());
  vector<double> vars(get_nComps());

  // Get the global variance
  getGlobalVar(vars);

  // Create the data structure to hold the normalized features
  vector<vector<float> > normedFeats(get_nObs());

  // Create the normalized features
  vector<float> tmpVec(get_nComps()); // tmp vector for the CMS features
  for(int i = 0; i<get_nObs(); i++) {
    //getMeanVar(winBounds[i],means,vars); // compute means and vars
    getMean(winBounds[i],means); // compute means and vars
    vector<float>& curVec = spOnly[i]; // get current feature vector
    for(int dim=0;dim<curVec.size();dim++) { 
      tmpVec[dim] = (curVec[dim] - means[dim]) / vars[dim]; // do normalization
    }
    // save the normalized feature vector
    normedFeats[i].assign(tmpVec.begin(),tmpVec.end());	
  }

  // Replace the unnormalized features with the normalized features
  spOnly.assign(normedFeats.begin(),normedFeats.end());
}

void
HTKfile::getMean(windowBounds wb, vector<double>& m) {

  m.assign(get_nComps(),0.0);

  // Compute the sum along each dimension of the feature vector
  int fvIndex = wb.getBeg();
  int wEnd    = wb.getEnd();
  for( ; fvIndex <= wEnd; fvIndex++) {
    for(int i = 0; i<=get_nComps(); i++) {
      m[i] += spOnly[fvIndex][i];
    }
  }

  double N = static_cast<double>(wb.getEnd()-wb.getBeg()+1);
  for(int i = 0; i<get_nComps(); i++) {
    m[i] /= N;
  }  

}

void
HTKfile::getMeanVar(windowBounds wb, vector<double>& m, vector<double>& v) {

  vector<double> tmpM(m.size());
  vector<double> tmpV(v.size());

  // Compute the sum along each dimension of the feature vector
  int fvIndex = wb.getBeg();
  int wEnd    = wb.getEnd();
  for( ; fvIndex <= wEnd; fvIndex++) {
    for(int i = 0; i<=get_nComps(); i++) {
      float tmp = spOnly[fvIndex][i];
      tmpM[i] += tmp;
      tmpV[i] += tmp*tmp;
    }
  }

  double N = static_cast<double>(wb.getEnd()-wb.getBeg()+1);
  for(int i = 0; i<get_nComps(); i++) {
    tmpM[i] /= N;
    tmpV[i] /= N-1.0;
    tmpV[i] -= (N * tmpM[i] * tmpM[i])/(N-1.0);
    if (tmpV[i] < MIN_VARIANCE) tmpV[i] = MIN_VARIANCE;
  }  

  m.assign(tmpM.begin(),tmpM.end());
  v.assign(tmpV.begin(),tmpV.end());
}

void
HTKfile::getGlobalVar(vector<double>& v) {
  // tmp vector to hold the means, which we throw away
  vector<double> tmpM(v.size());
  // create a windowBounds that covers all of the feature vectors
  windowBounds wb(0,get_nObs()-1);
  // get the variance
  getMeanVar(wb,tmpM,v);
}

void
HTKfile::getWinBounds(vector<windowBounds>& bnds, unsigned int winlen) {
  int nobs = get_nObs();
  bnds.resize(nobs);		// make sure there is enough room
  for (int i=0;i<nobs;i++) {
    bnds[i].setBounds(i,winlen,nobs-1);
  }
}

// Return the number of "working" frames. If the non-speech frames
// have been removed, then this will return the number of speech-only
// frames.
int 
HTKfile::get_nObs() const {
  int n;
  if (spOnlyFlag) 
    n = nSpOnly;
  else
    n = info.nObs;

  return n;
}

// Return the number of frames in the *original* file (the file
// as it was read from disk.)
int 
HTKfile::get_nObsOrig() const {
  return info.nObs;
}

//
// Read in a speech segments file. This file is expected to be formatted
// just like a UEM file.
//
void
HTKfile::readUEM(const string fname) {
  if(!openDone) {
    cerr << "Attempt to call HTKfile::readSpeechSegs() before HTKfile::open()"
	 << endl;
    exit(1);
  }

  cout << "Extracting speech-only frames." << endl;
  // Compute the number of seconds per frame (have to divide by 1000
  // because get_samplePeriod() returns milliseconds.)
  float secPerFrame = get_samplePeriod()/1000.0;

  // This will hold the start,end pairs
  vector<pair<float,float> > segvec;

  // Create the input stream
  ifstream ssinfile(fname.c_str(),ios::in);
  if (! ssinfile.is_open()) {
    cerr << "Error opening: " << fname << endl;
    exit(1);
  }

  // Read in the UEM data and store the start and end times in segvec
  float s,e;
  string showname;
  int chan;
  while(ssinfile >> showname >> chan >> s >> e){
    segvec.push_back(make_pair(s,e));
  }
  ssinfile.close();
  int nSpSegs = segvec.size();
  cout << "  Found " << nSpSegs << " speech segments in " 
       << fname << endl;

  // For each frame in the feature file, set a flag to TRUE if the frame
  // is a speech frame.
  // For each frame in the feature file, set a flag to TRUE if the frame
  // is a speech frame.
  int nframes = get_nObs();
  vector<bool> spflags(nframes,FALSE); // a vector of boolean flags
  for(int i=0;i<nSpSegs;i++) {
    float s = floor(100 * segvec[i].first+0.5); //we do a round in the  10ms time
    float e = floor(100 * segvec[i].second+0.5);

    // calculate the frame index for the start time
    //be careful that floating point is not exact XAM
    int si = int(floor((s / (secPerFrame*100))));
    // calculate the frame index for the end time
    int ei = int(floor((e / (secPerFrame*100))));

    printf("Speech from %d to %d (%.10f, %.10f)(%.10f, %.10f)\n", si,
	   ei,s, e, segvec[i].first,segvec[i].second );

    if (si < nframes) {
      if (ei >= nframes) ei = nframes - 1;
      for (int j = si; j<=ei; j++){
	spflags[j] = TRUE;
	nSpOnly++;		// class variable
      }
    }
  }

  // Next, fill the vector of speech-only frames.
  fillSPOnly(spflags);

  cout << "  Found " << nSpOnly << " frames of speech." << endl;
  cout << "Done." << endl;
}

// Fill the vector of speech-only frames 'spOnly' based on the bool
// vector of speech flags. 
// (The 'spOnly' vector is where readObs() should return data from.)
// Also, create a map that shows us which frames in the speech-only
// data map to which frames in the original file.
// The 'spflags' vector is assumed to contain one flag for each speech
// vector in the file.
void
HTKfile::fillSPOnly(vector<bool>& spflags) {
  int j = 0;
  int ncomps = get_nComps();	// number of features per frame
  for(int i=0;i<spflags.size();i++){
    if(spflags[i]){		// is this a speech frame?
      vector<float> curframe(ncomps,0.0);
      readObs(i,&curframe[0]); // read this frame of data
      spOnly.push_back(curframe); // add data to the speech-only vector
      frameMap.insert(make_pair(j,i)); // insert this frame into the map
      j++;
    }
  }

  // set the class var 'nSpOnly' that contains the number of speech-only frames
  nSpOnly = spOnly.size();
  spOnlyFlag = TRUE;
}

int
HTKfile::mapFrame(const int i) const {
  if (spOnlyFlag) {
    map<int,int>::const_iterator pos;
    pos = frameMap.find(i);
    if (pos != frameMap.end())
      return pos->second;
    else {
      cerr << "HTKfile::getFrameMap: frame " << i << " not in frameMap." << endl;
      exit(1);
    }
  } else {			// no mapping to be done
    int j = i;
    return j;
  }
}

void
HTKfile::init(){
 vaxOrder=IsVAXOrder();
 InitMem();
 InitParm();

 // Allocate memory for the file
 CreateHeap(&x,"BufferIn",MSTAK,1, 0.5, 100000, 5000000);

 initDone = TRUE;
}

// IsVAXOrder: returns true if machine has VAX ordered bytes
// This routine was taken directly from HShell.c in the HTKLib subdir.
Boolean HTKfile::IsVAXOrder(void)
{
   short x, *px;
   unsigned char *pc;

   px = &x;
   pc = (unsigned char *) px;
   *pc = 1; *(pc+1) = 0;         /* store bytes 1 0 */
   return (x==1 ? TRUE : FALSE);          /* does it read back as 1? */
}

void
HTKfile::printinfo() const {
  if(!openDone) {
    cerr << "Attempt to call HTKfile::printinfo() before HTKfile::open()"
	 << endl;
    exit(1);
  }

  char buf[MAXSTRLEN];

  // number of bytes per observation (frame)
  printf("  Sample Bytes:  %-7d", info.srcVecSize*sizeof(float));
  // the HTK sample kind (e.g. "LPCEPSTRA") 
  printf("  Sample Kind:   %s\n", ParmKind2Str(info.srcPK,buf));
  // number of components (features) per observation
  printf("  Num Comps:     %-7d", info.srcVecSize);
  // frame rate
  printf("  Sample Period: %.1f msecs\n", info.srcSampRate/10000.0);
  // total number of observations
  printf("  Num Obs:   %-7d", info.nObs);
  // the source file format (e.g. "HTK")
  printf("  File Format:   %s\n", Format2Str(info.srcFF));

}

// return the number of observations (feature vectors) in the file
int 
HTKfile::open(const char *fname) {
  if (!initDone) {
    cerr << "Attempt to call HTKfile::open() before HTKfile::init()"
	 << endl;
    exit(1);
  }

  short swidth[SMAX];
  Boolean eSep = FALSE;

  // assuming only one stream
  ZeroStreamWidths(1,swidth);

  // Open the HTK file and read the entire file into memory. 
  // If you don't want to read the entire file into memory, then you
  // should change the third argument to, say, 50. This will cause
  // the file to be read in in "chunks", which is good if you are only
  // going to read through the file one time and you don't want to store
  // it all in memory. However, if you plan to do random access on the file
  // the leave the third argument set at '0'. See the file HParm.h in the
  // HTKLib directory for more info.
  //
  if((pbuf = OpenBuffer(&x,const_cast<char*>(fname),0,UNDEFF,TRI_UNDEF,
			TRI_UNDEF))==NULL) {
    return(0);
  }

  // In addition to getting info about the buffer, it also reads
  // the entire file into memory.
  GetBufferInfo(pbuf,&info);

  SetStreamWidths(info.tgtPK,info.tgtVecSize,swidth,&eSep);

  o = MakeObservation(&gstack,swidth,info.tgtPK,
		      static_cast<Boolean>(FALSE),eSep);

  StartBuffer(pbuf);  

  // run through all of the data looking for invalid numbers
  for(int i=0;i<info.nObs;i++) {
    ReadAsTable(pbuf,i,&o);

    for(int j = 1;j<=VectorSize(o.fv[1]);j++) {
      if(isnan(o.fv[1][j]) || !finite(o.fv[1][j])) {
	printf("data at frame %d not valid: %f\n",i,o.fv[1][j]);
	fflush(stdout);
	exit(1);
      }
    }
  }
  openDone = TRUE;
  return(info.nObs);
}

void
HTKfile::addChunkInfo(char* fname) { // reads chunk info in from file
  if(!openDone) {
    cerr << "Attempt to call HTKfile::addChunkInfo() before HTKfile::open()"
	 << endl;
    exit(1);
  }

  if(chunkinfo != static_cast<unsigned char*>(NULL)) {
    delete [] chunkinfo;
  }
  chunkinfo = new unsigned char[info.nObs];

  FILE *fp = fopen(fname,"r");
  if(fp == NULL){
    fprintf(stderr,"Error reading %s\n",fname);
    exit(0);
  }
  char buf[10];
  for(int i = 0;i<info.nObs;i++){
    if(fscanf(fp,"%d\n",(chunkinfo+i)) != 1) {
      fprintf(stderr,"Error reading %d th frame from %s\n",i,fname);
      exit(0);
    }
  }
  fclose(fp);
}

unsigned char
HTKfile::getChunkInfo(const int frame) const {
  if(chunkinfo == static_cast<unsigned char*>(NULL))
    return 0;			// no chunk info available

  if (frame >= info.nObs || frame < 0)
    return 0;			// frame out of range

  return chunkinfo[frame];
}

void
HTKfile::readObs(int i, float* output) const {
  if (!openDone) {
    cerr << "Attempt to call HTKfile::readObs() before HTKfile::open()"
	 << endl;
    exit(1);
  }

  // This is where we actually read from the HTK file. (Actually we
  // are just reading from memory. The disk file was read above in the
  // call to GetBufferInfo().)
  //
  // Since we called OpenBuffer (above) with a third arg of 0, we can
  // randomly access this file using the ReadAsTable() function. 

  if (!spOnlyFlag){
    ReadAsTable(pbuf,i,&o);
    for(int j = 1;j<=VectorSize(o.fv[1]);j++)
      output[j-1] = o.fv[1][j];
  } else {
    if(i>=nSpOnly){
      cerr << "Error: request for frame at index " << i << " out of range."
	   << endl;
      exit(1);
    }
    for(int j = 0;j<get_nComps();j++)
      output[j] = spOnly[i][j];
  }    

}
