//
// $Id: hmm.cpp,v 1.1.1.1 2005/05/06 01:15:31 xanguera Exp $
//
#include <iostream>
#include <fstream>
#include <algorithm>
#include <functional>
#include <vector>
#include <set>

#include "hmm.h"

const char* message1 = "Number of clusters: "; 
const int len1 = 20; // length 20
const char* message2 = "Cluster ";
const int len2 = 8;

using namespace std;

////////////////////////////////////////////////////////////
HMM::HMM(int number_of_classes, int M, int dim, int md)
{ // M = number of Gaussians
  nGauss = M;
  nClass = number_of_classes;	// number of clusters
  DIM = dim;			// DIMension of feature vector
  MIN_DURATION = md;	    // MINimum DURATION of segment (in frames)
  ALWAYS_POSSIBLE = 0; // flag to allow merging despite NP; initialize to false
  U = new int[nClass]; // indicates whether a cluster is Usable or not
  gmm = new GMM[nClass];	// Gaussian Mixture Model
  lkld = new float[nClass];	// ???
  NP = (int **)malloc(sizeof(int *)*nClass); // list of clusters Not Possible to merge
  *NP = (int *)malloc(sizeof(int)*nClass*nClass); // NP[cd][0] = number of pairs, NP[cd][n] = other member of pair
  compared = (float **)malloc(sizeof(float *)*nClass); //holds the scores already computed in the neightbours segmentation
  *compared = (float *)malloc(sizeof(float)*nClass*nClass);
  for(int i=1;i<nClass;i++) {
    NP[i] = NP[i-1] + nClass;
    compared[i] = compared[i-1] + nClass;
  }
  for(int i=0; i<nClass; i++) {
    NP[i][0] = 0;  //the first element is going to indicate the number
    gmm[i].Initialize(M, DIM);
    for(int j=0; j<nClass; j++)
      {
	compared[i][j] = 0; //all elements are initially 0
      }
    U[i] = 1;
  }
}

////////////////////////////////////////////////////////////
HMM::~HMM()
{
  delete []U;
  delete []gmm;
  delete []lkld;
  free(*NP);
  free(NP);
}

////////////////////////////////////////////////////////////
HMM::HMM(){}

////////////////////////////////////////////////////////////
void 
HMM::InitializeXBIC(const HTKfile& htkf, chunk **sent, int nClus)
{
  int nSamples = htkf.get_nObs();

  // Get the number of seconds per frame (have to divide by 1000
  // because get_samplePeriod() returns milliseconds.)
  float secPerFrame = htkf.get_samplePeriod()/1000.0;

  float sLenSecs = 4.0;		// Length (in seconds) of the 2 xBIC segments
  int sLen = int(floor(sLenSecs/secPerFrame));

  float minsegsizeSecs = 2.0;	// Minimum segment size in seconds
  int minsegsize = int(floor(minsegsizeSecs/secPerFrame));

  float xBICstepSecs = 1.0;	// Step size (in seconds) to use during xBIC
  int xBICstep = int(floor(xBICstepSecs/secPerFrame));

  int nComps=9;			// Use this many gaussians per GMM

  printf("Total number of samples = %d (%0.1f minutes)\n", 
	 nSamples,static_cast<float>(nSamples)*secPerFrame/60.0);
  printf("Starting xBIC initialization\n");
  printf("  target is %d initial clusters.\n",nClus);
  printf("  xBIC GMMs have length %0.1f secs with %d gaussian comps.\n",
	 sLenSecs,nComps);
  printf("  xBIC step size is %0.1f secs.\n", xBICstepSecs);
  printf("  min cluster size will be %0.1f secs\n",minsegsizeSecs);

  // This will hold the xBIC scores for each frame
  // Typedef for convenience.
  typedef vector<float, allocator<float> > myvec;
  myvec xbicscores;

  // to store the indexes of the points on which we calculate the xbic
  // scores
  vector<int> xbicFno;
  // keep track of how many xbic scores we compute
  int xbicpoints = 0;

  int c1Start = 0;
  int c2Start = c1Start+sLen;

  // These are the two GMMs we will use in the xBIC comparisons
  GMM clus1, clus2;
  clus1.Initialize(nComps,DIM);
  clus2.Initialize(nComps,DIM);

  float *data1 = new float[sLen * DIM];	// data for GMM 1
  float *data2 = new float[sLen * DIM];	// data for GMM 2

  for(int fNo = sLen; fNo <= nSamples-sLen; fNo+=xBICstep) {
    // Read in the data for each of the GMMs to train on
    for(int i = 0; i<sLen; i++) {
      htkf.readObs(fNo-sLen+i,&data1[i*DIM]);
      htkf.readObs(fNo+i,&data2[i*DIM]);
    }
    // Train the GMMs
    clus1.Train(data1,sLen);
    clus2.Train(data2,sLen);

    // Get the xBIC score for this frame
    xbicscores.push_back(clus1.Evaluate(data2,sLen)+clus2.Evaluate(data1,sLen));
    // save this frame number
    xbicFno.push_back(fNo);	
    //printf("xBIC %d %f\n",fNo,xbicscores.back());
  }

  printf("  Done computing xBIC scores\n");

  // Find the min and max values of all of the xbic scores
  myvec::iterator minind, maxind;
  minind = min_element (xbicscores.begin(), xbicscores.end());
  maxind = max_element (xbicscores.begin(), xbicscores.end());
  float xbmin = *minind;
  float xbmax = *maxind;

  // Compute the step size for our grid search
  float stepsize = (xbmax - xbmin) / 100.0;

  // Create an set to hold the indices of the change points
  typedef set<int> mymap;
  mymap changePoints;

  // Find the threshold that will give us roughly nClus clusters
  float thresh = xbmin;
  myvec::size_type nelems = 0;
  while(thresh < xbmax && nelems < nClus-1){
    myvec::size_type candidates;

    thresh += stepsize;
    // Count all elements greater than the threshold.
    myvec::iterator it;
    it = find_if(xbicscores.begin(),xbicscores.end(),
		 bind2nd(less<myvec::value_type>(), thresh));

    // Find all elements that are less than the threshold (and
    // that occur at a local minimum)
    // and add them to the map. Only add them if there is not
    // already an index in the map that is within minsegsize frames
    // of the one we want to add.
    while( it != xbicscores.end() && changePoints.size() < nClus-1){
      // make sure this frame is not in the list of changePoints already
      myvec::difference_type offset = distance(xbicscores.begin(),it);
      if (offset < 0) offset *= -1;
      //int xbfi = static_cast<int>(offset);
      int findex = xbicFno[offset];	// frame index of this xbic score
      if(changePoints.count(findex) == 0) {
	// are we at a local minimum?
	if ((it != xbicscores.begin()) && *it < *(it-1) 
	    && ((it+1) != xbicscores.end()) && *it < *(it+1)) {
	  int found = 0;
	  mymap::iterator mapi = changePoints.begin();
	  while(mapi != changePoints.end()){
	    // how close are we to already stored frame indexes?
	    int dist = findex - *mapi; 
	    if (dist < 0) dist *= -1;
	    if (dist < minsegsize) {
	      found = 1;	// consider it already in the map
	      break;
	    }
	    mapi++;
	  }
	  if (! found) {	// this change point was not found
	    changePoints.insert(findex);
	  }
	}
      }
      it++;
      // find the next value that is less than thresh
      it = find_if(it,xbicscores.end(),
		   bind2nd(less<myvec::value_type>(), thresh));
    }
    nelems = changePoints.size();
    //printf("%d values less than %f\n",static_cast<int>(nelems),thresh);
  }
  printf("  Found %d initial clusters:\n",nelems+1);

  // Now populate the 'sent' datastructure using the values in the 
  // changePoints set
  *sent = new chunk;
  chunk *curr = *sent;
  curr->from = 0;
  int bicind = 1;
  int j = 0;

  mymap::iterator mapi = changePoints.begin();
  while(mapi != changePoints.end()) {
    int ind = *mapi;
    curr->cluster = j;
    curr->to = ind - 1;
    printf("    %d: %d to %d (%d frames)\n",j,curr->from, curr->to, 
	   (curr->to - curr->from + 1));
    curr->next = new chunk;
    curr->next->prev = curr;
    curr->next->from = ind;
    curr = curr->next;
    j++;
    mapi++;
  }

  // Finish off the final cluster
  curr->cluster = j;
  curr->next = NULL;
  curr->to = nSamples-1;
  printf("    %d: %d to %d (%d frames)\n",j,curr->from, curr->to, 
	 (curr->to - curr->from));

  // Initialize GMMs with these clusters
  float *data = new float[nSamples * DIM];
  for(int cl=0; cl<nClass;cl++){
    // count = number of frames in cluster cl
    int count = Get_Nframes(htkf, *sent, cl, cl);
    Get_Data_noalloc(htkf, data, *sent, cl, cl);
    gmm[cl].Initialize(data, count);
  }
  delete [] data;
  delete [] data1;
  delete [] data2;
}

////////////////////////////////////////////////////////////
void 
HMM::Initialize(const HTKfile& htkf, chunk **sent, const char* const fname)
{
  int nSamples = htkf.get_nObs();
  printf("Total number of samples = %d\n", nSamples);

  // Compute the number of seconds per frame (have to divide by 1000
  // because get_samplePeriod() returns milliseconds.)
  float secPerFrame = htkf.get_samplePeriod()/1000.0;

  // This will hold the start,end pairs
  vector<pair<float,float> > segvec;

  // Create the input stream
  ifstream ssinfile(fname,ios::in);
  if (! ssinfile.is_open()) {
    cerr << "Error opening: " << fname << endl;
    exit(1);
  }

  float s,e;
  while(ssinfile >> s >> e){
    segvec.push_back(make_pair(s,e));
  }
  ssinfile.close();
  int nSegs = segvec.size();

  if(nSegs != nClass) {
    cerr << "Initialization error: number of segments in " << fname <<
      " does not equal number of initial segs (" << nClass << ")" << endl;
    exit(1);
  }
  cout << "  Found " << nSegs << " initial segments in " 
       << fname << endl;

  // Create the sent structure
  *sent = new chunk;
  chunk *curr = *sent;

  for(int i=0;i<nSegs;i++) {
    s = segvec[i].first;
    e = segvec[i].second;

    // calculate the frame index for the start time
    int si = int(floor(s / secPerFrame));
    // calculate the frame index for the end time 
    int ei = int(floor(e / secPerFrame));

    curr->cluster = i;
    curr->from    = si;
    curr->to      = ei;
    cout << i << ": " << si << " to " << ei << 
      " (" << ei-si << " frames)" << endl;
    if (i == nSegs-1){		// last cluster
      curr->next = NULL;
    } else {
      curr->next    = new chunk;
      curr->next->prev = curr;
      curr = curr->next;
    }
  }

  // figure out how man guassians there are for the entire file
  int totgauss = nGauss * nClass;
  // Initialize GMMs using the cluster boundaries we just read in
  float *data = new float[nSamples * DIM];
  for(int cl=0; cl<nClass;cl++){
    // count = number of frames in cluster cl
    int count = Get_Nframes(htkf, *sent, cl, cl); 
    // figure out how many gaussians should be allocated to this
    // cluster
    float framefrac = static_cast<float>(count)/static_cast<float>(nSamples);
    int cl_nGauss = int(floor(framefrac*static_cast<float>(totgauss)));
    gmm[cl].Initialize(cl_nGauss,DIM); // reset number of gaussians
    cout << "Cluster " << cl << " getting " << cl_nGauss << " gaussians."
	 << endl;
    Get_Data_noalloc(htkf, data, *sent, cl, cl);
    gmm[cl].Initialize(data, count);
  }
  delete [] data;
}

////////////////////////////////////////////////////////////
void 
HMM::Initialize(const HTKfile& htkf, chunk **sent)
{
  int nSamples = htkf.get_nObs();
  printf("Total number of samples = %d\n", nSamples);

  int gap = nSamples/(nClass*P); // P is defined in global.h (=1)

  // Uniform segmentation with gap frames per cluster
  *sent = new chunk;
  chunk *curr = *sent;
  curr->from = 0;
  for(int i=0;i<P;i++){                           // P == 2
    for(int j=0;j<nClass;j++){
      curr->cluster = j;
      if(i==P-1 && j==nClass-1){                  // Last cluster
	curr->next = NULL;
	curr->to = nSamples-1;
      }else{
	curr->to = curr->from + gap -1;
	curr->next = new chunk;
	curr->next->prev = curr;
	curr = curr->next;
	curr->from = curr->prev->to+1;
      }
    }
  }

  // Initialize GMMs with these clusters
  float *data = new float[nSamples * DIM];
  for(int cl=0; cl<nClass;cl++){
    int count = Get_Nframes(htkf, *sent, cl, cl); // number of frames in cluster cl
    Get_Data_noalloc(htkf, data, *sent, cl, cl);
    gmm[cl].Initialize(data, count); // shouldn't the 2nd parameter be DIM?
  }
  delete [] data;
}

#ifdef SEGMENT_TIMER
#include <time.h>
double HMM_time()
{
      static double starttime=0;
      double now;
      timeval tv;
      gettimeofday(&tv,NULL);
      now = (double) tv.tv_sec + ((double)tv.tv_usec)/1.0e6;
      if (!starttime) starttime = now;
      return now-starttime;
}
#endif

////////////////////////////////////////////////////////////
float 
HMM::segment(const HTKfile& htkf, chunk **sent) {
  static int nSamples = htkf.get_nObs();
  static float *data;
  static int *F;
  static int *R;
  static double **dC;
  static int **tC;
  static int memInitDone = 0;
  // (MIN_DURATION*nClass) = total number of states in all models (classes)
  int totStates = MIN_DURATION*nClass;
  if (!memInitDone){
    F = new int[nSamples];	// hold the frame backpointer
    R = new int[nSamples];	// hold the best cluster at each frame

    dC = new double*[2];		// hold the best scores for each state
    dC[LAST] = new double[totStates];
    dC[CURR] = new double[totStates];

    tC = new int*[2];		// hold the frame pointers for each state
    tC[LAST] = new int[totStates];
    tC[CURR] = new int[totStates];

    data = new float[DIM];
    memInitDone = 1;
  }
  F[0] = 0;
  R[0] = 0;

  int n_usable = 0;
  for(int cl=0; cl<nClass; cl++){
    if(U[cl])
      n_usable++;
  }
  double TRANS_PROB = log(0.1/static_cast<double>(n_usable));
 
  printf("Beginning segmentation, nSamples = %d, nClass = %d\n", nSamples, 
	 n_usable);
  fflush(stdout);

  // first initialize the global scores for all states to a small likelihood
  for(int st=0; st<totStates; st++){
    dC[LAST][st] = MINVALUEFORMINUSLOG;
    dC[CURR][st] = MINVALUEFORMINUSLOG;
    tC[LAST][st] = 0;
    tC[CURR][st] = 0;
  }

  // now read in the data for the 0th frame and set state 0 of each model
  // to the likelihood of the data given the model.
  htkf.readObs(0,data);
  Calculate_Lkld(data);
  //for(int cl=0; cl<nClass; cl++){
    //printf("lkld-%d: %f\n",cl,lkld[cl]);
    //lkld[cl] += 1000;//+=19*log(2*3.141592654)/2;
  //}

  for(int cl=0; cl<nClass;cl++)
    if(U[cl]) {
      dC[LAST][(cl*MIN_DURATION)] = (double)(lkld[cl]);
      dC[CURR][(cl*MIN_DURATION)] = (double)(lkld[cl]);
    }

  // 
  // Loop over remaining frames of data
  //
  for(int frame = 1; frame <nSamples; frame++) {
    htkf.readObs(frame,data);

    Calculate_Lkld(data);
    //for(int cl=0; cl<nClass; cl++){
      //printf("lkld-%d: %f\n",cl,lkld[cl]);
      //lkld[cl] += 1000;//+=19*log(2*3.141592654)/2;
    //}

    // 
    // Handle state 0 of each model
    //   This state *must* come from the final state of best model
    //   at the previous time.
    //

    // the index of the final state of the best previous cluster
    int bestGlobalState = R[frame-1]*MIN_DURATION + MIN_DURATION-1;
    // the cost of transitioning from the best previous cluster
    double clusterTrans = dC[LAST][bestGlobalState] + TRANS_PROB;
    //printf("Best prev. lkld (%d): %f\n", frame, clusterTrans);

    for(int cl=0; cl<nClass; cl++){
      int curState;
      if(U[cl]){
	//
	// Get the index of the 0th state of current cluster
	//
	curState = cl*MIN_DURATION;	

	// add local distance to state's global score
	dC[CURR][curState] = clusterTrans + (double)(lkld[cl]);
	// since we are transitioning now, this frame is the begin time
	tC[CURR][curState] = frame;
      }
    }

    // 
    // Handle the other states of each cluster
    //
    for(int cl=0; cl<nClass;cl++) {
      double forwardTrans;
      int gl_state;		// index into global state array
      int cl_state;		// index of state in a cluster
      if(U[cl]) {
	double localDist = (double)(lkld[cl]);

	//
	// Start with state 1 and end at the second-to-last state
	//   There are no self-loops on the internal states!
	//
	for(cl_state = 1; cl_state < (MIN_DURATION-1); cl_state++) {
	  gl_state = (cl*MIN_DURATION) + cl_state;
	  forwardTrans = dC[LAST][gl_state-1];
	  dC[CURR][gl_state] = forwardTrans + localDist;
	  tC[CURR][gl_state] = tC[LAST][gl_state-1];
	}

	//
	// Now do the final state of each cluster
	//  This is handled differently from the internal states because
	//  it has a self-loop prob of 0.9
	//
	gl_state = (cl*MIN_DURATION) + cl_state;
	double selfLoopTrans = dC[LAST][gl_state] + LOG_0_9;
	forwardTrans = dC[LAST][gl_state-1];
	//printf("Difference %d - %d: %.15f (%.15f - %.15f)\n", frame, cl, forwardTrans - selfLoopTrans, forwardTrans, selfLoopTrans );
	if (selfLoopTrans > forwardTrans) {
	  dC[CURR][gl_state] = selfLoopTrans + localDist;
	  tC[CURR][gl_state] = tC[LAST][gl_state];
	} else {
	  dC[CURR][gl_state] = forwardTrans + localDist;
	  tC[CURR][gl_state] = tC[LAST][gl_state-1];
	}
      }
    }

    // 
    // Finished with all states of all models. Now find the cluster
    // that has the highest likelihood in its final state.
    //
    double MaxLkld = MIN_FLOAT;
    int MaxClass = -1;
    int MaxClassFinSt = -1;
    for(int cl=0; cl<nClass; cl++) {
      if(U[cl]) {
	int finalClusState = cl*MIN_DURATION + MIN_DURATION-1;
	double tmplkld = dC[CURR][finalClusState];
	if(tmplkld > MaxLkld) {
	  MaxClass = cl;
	  MaxLkld = tmplkld;
	  MaxClassFinSt = finalClusState;
	}
      }
    }
    //printf("Best class at %d: %f\n", frame, MaxLkld);
    if (MaxClass == -1) {
      printf("HMM::segment() - error finding best class\n");
      exit(1);
    }
    // Record the results
    R[frame] = MaxClass;
    F[frame] = tC[CURR][MaxClassFinSt];

    // copy CURR to LAST
    for(int st=0;st <totStates; st++) {
      dC[LAST][st] = dC[CURR][st];
      tC[LAST][st] = tC[CURR][st];
    }

  } /* frame loop */

  // create linked list from back to front  
  *sent = new chunk;
  (*sent)->prev = NULL;		// not yet created
  (*sent)->next = NULL;		// will remain NULL, since this is the last
  (*sent)->cluster = R[nSamples-1]; // cluster number
  (*sent)->from = F[nSamples-1]; // start time
  (*sent)->to = nSamples-1;	// end time
  while((*sent)->from > 1) {	// stop when we get to beginning of utterance
    (*sent)->prev = new chunk;
    (*sent)->prev->next = (*sent);

    (*sent) = (*sent)->prev;	// point at newly created chunk
    (*sent)->to = (*sent)->next->from - 1; // end == next chunk's start - 1
    (*sent)->from = F[(*sent)->to]; // look up start time in F[]
    (*sent)->cluster = R[(*sent)->to]; // look up best cluster in R[]
    (*sent)->prev = NULL;
  }
  int bestFinalClus = R[nSamples-1];
  int bestFinalState = bestFinalClus * MIN_DURATION + MIN_DURATION - 1;
  float viterbi_score = (float)(dC[CURR][bestFinalState]); //I cast back from double to float
  printf("   Segmentation Over with Viterbi score = %f\n", viterbi_score);
  fflush(stdout);

  return viterbi_score;
}

////////////////////////////////////////////////////////////
void 
HMM::Calculate_Lkld(float *vector) {
  for(int cl=0; cl<nClass; cl++) {
    if(U[cl])
      lkld[cl] = gmm[cl].Log_Likelihood(vector);
  }
}

////////////////////////////////////////////////////////////
void 
HMM::Train_HMM(const HTKfile& htkf, chunk *first){
  for(int cl=0; cl<nClass; cl++){
    if(U[cl]){
      int count = Get_Nframes(htkf, first, cl, cl);  // number of frames in cluster cl
      if(count > gmm[cl].M) {                        // Why compare to number of Gaussians?
	float *data = new float[count*DIM];
	Get_Data_noalloc(htkf, data, first, cl, cl);
	gmm[cl].lkld = gmm[cl].Train(data, count);   // 
	//printf("lkld %d %f %f\n",cl,gmm[cl].lkld,gmm[cl].lkld/(float)count);
	delete [] data;
      } else
	U[cl] = 0;
    }
  }
}

////////////////////////////////////////////////////////////
void 
HMM::Train_HMM_Individual(const HTKfile& htkf, chunk *first, int cl){
  if(U[cl]){
    int count = Get_Nframes(htkf, first, cl, cl);  // number of frames in cluster cl
    if(count > gmm[cl].M) {                        // Why compare to number of Gaussians?
      float *data = new float[count*DIM];
      Get_Data_noalloc(htkf, data, first, cl, cl);
      gmm[cl].lkld = gmm[cl].Train(data, count);   // 
      delete [] data;
    } else
      U[cl] = 0;
  }
}

////////////////////////////////////////////////////////////
//
// Count the number of frames assigned to the two chunks
//
int
HMM::Get_Nframes(const HTKfile& htkf, chunk *first, int cd1, int cd2) {

  int fCount = 0;

  // count the total number of frames assigned to the 
  // two clusters- cd1 and cd2
  for(chunk *ch = first; ch!=NULL; ch= ch->next){
    if(ch->cluster == cd1 || ch->cluster == cd2){
      fCount += (ch->to - ch->from + 1);
    }
  }
  return fCount;
}

////////////////////////////////////////////////////////////
void 
HMM::Get_Data(const HTKfile& htkf, float ***pData, chunk *first, 
	      int cd1, int cd2, int *pCount) {

  float **tmpDatap;
  float *tmpData;
  int fCount = 0;

  // count the total number of frames assigned to the 
  // two clusters- cd1 and cd2
  for(chunk *ch = first; ch!=NULL; ch= ch->next){
    if(ch->cluster == cd1 || ch->cluster == cd2){
      fCount += (ch->to - ch->from + 1);
    }
  }

  // create a block of memory to hold all of the data
  tmpData = (float *) malloc(sizeof(float*)*fCount*DIM);
  if(tmpData == NULL) {
    printf("Memory could not be allocated\n");
    exit(0);
  }
  // now create a bunch of pointers into the block of memory above
  tmpDatap = (float **) malloc(sizeof(float**)*fCount);
  if(tmpDatap == NULL){
    printf("Memory could not be allocated\n");
    exit(0);
  }
  for(int i=0;i<fCount;i++)
    tmpDatap[i] = &tmpData[i*DIM];

  for(chunk *ch = first; ch!=NULL; ch= ch->next){
    if(ch->cluster == cd1 || ch->cluster == cd2){
      int index = 0;
      for(int smp=ch->from; smp <= ch->to; smp++){
	htkf.readObs(smp,tmpDatap[index]);
	index++;
      }/*for(int smp=ch->from; smp < ch->to; smp++)*/
    }/*if(ch->cluster == cd1 || ch->cluster == cd2)*/
  }/*for(chunk *ch = *pFirst; ch!=NULL; ch= ch->next)*/

  *pCount = fCount;
  *pData = tmpDatap;
}

////////////////////////////////////////////////////////////
//
// Version of Get_Data() that assumes that the passed-in array has
// enough space to hold the requested data.
//
void 
HMM::Get_Data_noalloc(const HTKfile& htkf, float *pData, chunk *first, 
	      int cd1, int cd2) {

  int index = 0;
  for(chunk *ch = first; ch!=NULL; ch= ch->next){
    if(ch->cluster == cd1 || ch->cluster == cd2){
      for(int smp=ch->from; smp <= ch->to; smp++){
	htkf.readObs(smp,(pData+(index*DIM)));
	index++;
      }/*for(int smp=ch->from; smp < ch->to; smp++)*/
    }/*if(ch->cluster == cd1 || ch->cluster == cd2)*/
  }/*for(chunk *ch = *pFirst; ch!=NULL; ch= ch->next)*/
}

////////////////////////////////////////////////////////////
int 
HMM::Find_Best(int *mcd1, int *mcd2, chunk *first, const HTKfile& htkf) {
  int count;
  int FOUND = 0;
  float thresh_diff = 0.0;
  float merge_lkld, max_diff = -100000.0;
  float *data;
  for(int cd1=0; cd1 < nClass-1; cd1++) {
    for(int cd2=cd1+1; cd2 < nClass; cd2++) {
      if(U[cd1] && U[cd2] && if_possible(cd1, cd2)) {
	//printf("Testing: %d %d: ",cd1,cd2);
	count = Get_Nframes(htkf,first,cd1,cd2);
	data = new float[count*DIM];
	Get_Data_noalloc(htkf, data, first, cd1, cd2);

	int cd1_M = gmm[cd1].M;	// number of mixtures in cluster 1
	int cd2_M = gmm[cd2].M;	// number of mixtures in cluster 2
	int nComps = gmm[cd1].M + gmm[cd2].M; // num new mixtures

	// Compute the ratio of the number of mixtures in each 
	// of the two clusters to the total number of new mixtures.
	// This tells us how much we should weight each of the 
	// individual clusters when creating the combined cluster.
	float r1 = static_cast<float>(cd1_M)/static_cast<float>(nComps);
	float r2 = static_cast<float>(cd2_M)/static_cast<float>(nComps);

	GMM temp_gmm;
	temp_gmm.Initialize(nComps, DIM);
	for(int i=0; i<cd1_M; i++) {
	  temp_gmm.Weight[i] =gmm[cd1].Weight[i]*r1;
	  memcpy(temp_gmm.gArray[i].Mean, gmm[cd1].gArray[i].Mean, 
		 sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i].Var, gmm[cd1].gArray[i].Var, 
		 sizeof(float)*DIM);
	}
	for(int i=0; i<cd2_M; i++) {
	  temp_gmm.Weight[i+cd1_M] = gmm[cd2].Weight[i]*r2;
	  memcpy(temp_gmm.gArray[i+cd1_M].Mean, gmm[cd2].gArray[i].Mean, 
		 sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i+cd1_M].Var, gmm[cd2].gArray[i].Var, 
		 sizeof(float)*DIM);
	}
	temp_gmm.lkld = temp_gmm.Train(data, count);
	float temp_diff = temp_gmm.lkld - (gmm[cd1].lkld + gmm[cd2].lkld);
	//printf("%f\n",temp_diff);
	if(temp_diff > thresh_diff && temp_diff > max_diff) {
	  max_diff = temp_diff;
	  *mcd1 = cd1;
	  *mcd2 = cd2;
	  FOUND = 1;
	}

	if(temp_diff < -100.0) {
	  NP[cd1][0]++;
	  NP[cd1][NP[cd1][0]] = cd2;
	  NP[cd2][0]++;
	  NP[cd2][NP[cd2][0]] = cd1;
	}

	delete [] data;
      }/*if(U[cd1] && U[cd2])*/
    }/*for(int cd2=cd1+1; cd2 < nClass; cd++)*/
  }/*for(int cd1=0; cd1 < nClass-1; cd1++)*/
  return FOUND;
}

////////////////////////////////////////////////////////////

int 
HMM::Find_Best_Adjacent(int *mcd1, int *mcd2, chunk *first, const HTKfile& htkf) {
  int count;
  int FOUND = 0;
  float merge_lkld, max_diff;
  float *data;
  int cd1,cd2;
  chunk *chunk_now;

  //amount of chunks active
  int nClusters = nU();
  
  max_diff = MIN_MERGE_SCORE;
  
  chunk_now = first;

  //In a first round we calculate the scores for the values not computed before
  while(chunk_now->next != NULL)
    {
      cd1 = chunk_now->cluster;
      cd2 = chunk_now->next->cluster;
      if(U[cd1] && U[cd2] && if_possible(cd1, cd2) && !compared[cd1][cd2]) {
	count = Get_Nframes(htkf,first,cd1,cd2);
	data = new float[count*DIM];
	Get_Data_noalloc(htkf, data, first, cd1, cd2);
	GMM temp_gmm;
	temp_gmm.Initialize(gmm[cd1].M+gmm[cd2].M, DIM);
	for(int i=0; i<gmm[cd1].M; i++) {
	  temp_gmm.Weight[i] =gmm[cd1].Weight[i]*(float)gmm[cd1].M/((float)gmm[cd1].M+gmm[cd2].M);
	  memcpy(temp_gmm.gArray[i].Mean, gmm[cd1].gArray[i].Mean, sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i].Var, gmm[cd1].gArray[i].Var, sizeof(float)*DIM);
	}
	for(int i=0; i<gmm[cd2].M; i++) {
	  temp_gmm.Weight[i+gmm[cd1].M] = gmm[cd2].Weight[i]*(float)gmm[cd2].M/((float)gmm[cd1].M+gmm[cd2].M);
	  memcpy(temp_gmm.gArray[i+gmm[cd1].M].Mean, gmm[cd2].gArray[i].Mean, sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i+gmm[cd1].M].Var, gmm[cd2].gArray[i].Var, sizeof(float)*DIM);
	}
	temp_gmm.lkld = temp_gmm.Train(data, count);
	
	float temp_diff = temp_gmm.lkld - (gmm[cd1].lkld + gmm[cd2].lkld);
	delete [] data;

	printf("Comparing %d with %d : %f\n",cd1,cd2,temp_diff);
	compared[cd1][cd2]=temp_diff;
	compared[cd2][cd1]=temp_diff;

      }/*if(U[cd1] && U[cd2])*/
      chunk_now = chunk_now->next; //next neighbour
    }/*while(chunk_now->next != NULL)*/

  //in a second pass we search for the best value
  printf("Second pass, looking for the best value\n");
  for(int cd1=0; cd1 < nClass-1; cd1++) {
    for(int cd2=cd1+1; cd2 < nClass; cd2++) {
      if(U[cd1] && U[cd2] && if_possible(cd1, cd2)) {    
	if(compared[cd1][cd2] > max_diff) {
	  max_diff = compared[cd1][cd2];
	  *mcd1 = cd1;
	  *mcd2 = cd2;
	  FOUND = 1;
	}
	if(compared[cd1][cd2] < 0.0) {
	  NP[cd1][0]++;
	  NP[cd1][NP[cd1][0]] = cd2;
	  NP[cd2][0]++;
	  NP[cd2][NP[cd2][0]] = cd1;
	}
      }
    }
  }
  
  return FOUND;
}

////////////////////////////////////////////////////////////
int 
HMM::Find_Best2(int *mcd1, int *mcd2, chunk *first, const HTKfile& htkf) {
  int count;
  //int FOUND = 0;
  int FOUND = 0, FIRST = 1;
  //float merge_lkld, max_diff = 0.0;
  float merge_lkld, max_diff;
  float *data;
  //printNP();
  for(int cd1=0; cd1 < nClass-1; cd1++) {
    for(int cd2=cd1+1; cd2 < nClass; cd2++) {
      if(U[cd1] && U[cd2] && if_possible(cd1, cd2)) {
	printf("Now testing clusters %d and %d.  ",cd1,cd2); fflush(stdout);
	count = Get_Nframes(htkf,first,cd1,cd2);
	data = new float[count*DIM];
	Get_Data_noalloc(htkf, data, first, cd1, cd2);
	GMM temp_gmm;
	temp_gmm.Initialize(gmm[cd1].M+gmm[cd2].M, DIM);
	for(int i=0; i<gmm[cd1].M; i++) {
	  temp_gmm.Weight[i] =gmm[cd1].Weight[i]*(float)gmm[cd1].M/((float)gmm[cd1].M+gmm[cd2].M);
	  memcpy(temp_gmm.gArray[i].Mean, gmm[cd1].gArray[i].Mean, sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i].Var, gmm[cd1].gArray[i].Var, sizeof(float)*DIM);
	}
	for(int i=0; i<gmm[cd2].M; i++) {
	  temp_gmm.Weight[i+gmm[cd1].M] = gmm[cd2].Weight[i]*(float)gmm[cd2].M/((float)gmm[cd1].M+gmm[cd2].M);
	  memcpy(temp_gmm.gArray[i+gmm[cd1].M].Mean, gmm[cd2].gArray[i].Mean, sizeof(float)*DIM);
	  memcpy(temp_gmm.gArray[i+gmm[cd1].M].Var, gmm[cd2].gArray[i].Var, sizeof(float)*DIM);
	}
	temp_gmm.lkld = temp_gmm.Train(data, count);

	float temp_diff = temp_gmm.lkld - (gmm[cd1].lkld + gmm[cd2].lkld);
	printf("Score = %f\n",temp_diff); fflush(stdout);
	/*if(temp_diff > max_diff) {
	  max_diff = temp_diff;
	  *mcd1 = cd1;
	  *mcd2 = cd2;
	  FOUND = 1;
	}*/
	if (FIRST == 1) {
	  max_diff = temp_diff;
	  *mcd1 = cd1;
	  *mcd2 = cd2;
	  FIRST = 0;
	} else {
	  if (temp_diff > max_diff) {
	    max_diff = temp_diff;
	    *mcd1 = cd1;
	    *mcd2 = cd2;
	  }
	}
	if(temp_diff < 0.0 && !ALWAYS_POSSIBLE) {
	  NP[cd1][0]++;
	  NP[cd1][NP[cd1][0]] = cd2;
	  NP[cd2][0]++;
	  NP[cd2][NP[cd2][0]] = cd1;
	}
	delete [] data;
      }/*if(U[cd1] && U[cd2])*/
    }/*for(int cd2=cd1+1; cd2 < nClass; cd++)*/
  }/*for(int cd1=0; cd1 < nClass-1; cd1++)*/

  if (max_diff > 0.0) {
    FOUND = 1;
  } else {
    FOUND = 0;
    ALWAYS_POSSIBLE = 1;
  }
  return FOUND;
}

////////////////////////////////////////////////////////////
int 
HMM::if_possible(int cd1, int cd2)
{
  if (ALWAYS_POSSIBLE) {
    return 1;
  }
  for(int i=1;i<=NP[cd1][0];i++) {
    if(NP[cd1][i] == cd2) {
      return 0;
    }
  }
  return 1;
}

////////////////////////////////////////////////////////////

void 
HMM::Merge(int cd1, int cd2, chunk *first, const HTKfile& htkf) {
  printf("Merging %d and %d\n",cd1,cd2);
  int count = Get_Nframes(htkf, first, cd1, cd2);
  float *data = new float[count*DIM];
  Get_Data_noalloc(htkf, data, first, cd1, cd2);
  GMM temp_gmm;
  temp_gmm.Initialize(gmm[cd1].M+gmm[cd2].M, DIM);
  for(int i=0; i<gmm[cd1].M; i++) {
    temp_gmm.Weight[i] =
      gmm[cd1].Weight[i]*(float)gmm[cd1].M/((float)gmm[cd1].M+gmm[cd2].M);
    memcpy(temp_gmm.gArray[i].Mean, gmm[cd1].gArray[i].Mean, sizeof(float)*DIM);
    memcpy(temp_gmm.gArray[i].Var, gmm[cd1].gArray[i].Var, sizeof(float)*DIM);
  }
  for(int i=0; i<gmm[cd2].M; i++) {
    temp_gmm.Weight[i+gmm[cd1].M]=
      gmm[cd2].Weight[i]*(float)gmm[cd2].M/((float)gmm[cd1].M+gmm[cd2].M);
    memcpy(temp_gmm.gArray[i+gmm[cd1].M].Mean, gmm[cd2].gArray[i].Mean, sizeof(float)*DIM);
    memcpy(temp_gmm.gArray[i+gmm[cd1].M].Var, gmm[cd2].gArray[i].Var,sizeof(float)*DIM);
  }
  temp_gmm.lkld = temp_gmm.Train(data, count);
  delete [] data;
  gmm[cd1].Copy(temp_gmm);
  U[cd2] = 0;
  clearNP(cd1); // allow cluster to be merged again
  //clearNP(cd2); // clean up NP
}
////////////////////////////////////////////////////////////

void 
HMM::Merge_Segmentation(int cd1, int cd2, chunk *first, const HTKfile& htkf) {
  chunk *chunk_now;
  int count = Get_Nframes(htkf, first, cd1, cd2);
  float *data = new float[count*DIM];
  Get_Data_noalloc(htkf, data, first, cd1, cd2);
  GMM temp_gmm;
  temp_gmm.Initialize(gmm[cd1].M+gmm[cd2].M, DIM);
  for(int i=0; i<gmm[cd1].M; i++) {
    temp_gmm.Weight[i] =
      gmm[cd1].Weight[i]*(float)gmm[cd1].M/((float)gmm[cd1].M+gmm[cd2].M);
    memcpy(temp_gmm.gArray[i].Mean, gmm[cd1].gArray[i].Mean, sizeof(float)*DIM);
    memcpy(temp_gmm.gArray[i].Var, gmm[cd1].gArray[i].Var, sizeof(float)*DIM);
  }
  for(int i=0; i<gmm[cd2].M; i++) {
    temp_gmm.Weight[i+gmm[cd1].M]=
      gmm[cd2].Weight[i]*(float)gmm[cd2].M/((float)gmm[cd1].M+gmm[cd2].M);
    memcpy(temp_gmm.gArray[i+gmm[cd1].M].Mean, gmm[cd2].gArray[i].Mean, sizeof(float)*DIM);
    memcpy(temp_gmm.gArray[i+gmm[cd1].M].Var, gmm[cd2].gArray[i].Var,sizeof(float)*DIM);
  }
  temp_gmm.lkld = temp_gmm.Train(data, count);
  delete [] data;
  gmm[cd1].Copy(temp_gmm);
  U[cd2] = 0;
  clearNP(cd1); // allow cluster to be merged again
  clearNP(cd2); // clean up NP

  //now we need to modify "first" to rename the cd2 chunks into cd1
  chunk_now = first;
  while(chunk_now->next != NULL)
    {
      if(chunk_now->cluster == cd2)
	{
	  chunk_now->cluster = cd1;
	}
	   

      chunk_now = chunk_now->next;
    }

}


////////////////////////////////////////////////////////////
void
HMM::clearNP(int cd) {
  NP[cd][0] = 0; // clear NP pairs starting with cd
  for (int cd1 = 0; cd1 < nClass; cd1++) { // for every possible first candidate
    for (int i = 1; i <= NP[cd1][0]; i++) { // for each listed pair
      if (NP[cd1][i] == cd) { // if (cd1,cd) pair found
	for (int j = i; j < NP[cd1][0]; j++) { // for each pair listed afterward
	  NP[cd1][j] = NP[cd1][j+1]; // shift over
	}
	NP[cd1][0]--; // decrement counter
      }
    }
  }
}

////////////////////////////////////////////////////////////
void
HMM::printNP() {
  for (int i = 0; i < nClass; i++) { // for each cluster
    if (U[i]) { // if cluster usable
      printf("%d: %d", i, NP[i][0]); // output cluster number and count
      for (int j = 1; j <= NP[i][0]; j++) { // for each pairing
	printf(" %d", NP[i][j]); // output other candidate
      }
      printf("\n"); // newline
    }
  }
  fflush(stdout);
}

////////////////////////////////////////////////////////////
void 
HMM::Write(char *fname, chunk *first) {
  FILE *fout = fopen(fname, "w");
  if(fout == NULL) {
    printf("Output file could not be opened\n");
    exit(0);
  }
  for(chunk *ch = first; ch!=NULL; ch=ch->next)
    fprintf(fout, "from %d to %d: cluster %d\n", ch->from + 1, ch->to + 1, 
	    ch->cluster);
  fclose(fout);
}

////////////////////////////////////////////////////////////
void 
HMM::Write(const char *fname, const chunk * const first, 
	   const HTKfile &htkf) {
  // First, run through the linked list of clusters and
  // create a new vector that has each frame individually labeled with its
  // cluster number. Nonspeech frames will be labeled with -1.
  int nframes = htkf.get_nObsOrig();
  vector<int> expanded(nframes,-1); // assume non-speech
  for(const chunk *ch = first; ch!=NULL; ch=ch->next)
    for(int i=ch->from;i<=ch->to;i++) {
      int oldfr = htkf.mapFrame(i); // get mapping to old frame number
      expanded[oldfr] = ch->cluster;
    }

  ofstream segout (fname);
  if (! segout.is_open()) {
    cerr << "Output file " << fname << "could not be opened" << endl;
    exit(0);
  }

  // Now go through the list of clusters and print it in seg.out format
  int cnum = expanded[0];
  int cst  = 0;
  int cen  = 0;
  for(int i=1;i<nframes;i++) {
    if (expanded[i] != cnum) {
      if (cnum < 0) {		// non-speech
	segout << "from " << cst+1 << " to " << cen+1 
	       << ": cluster <nonspeech>" << endl;
      } else {
	segout << "from " << cst+1 << " to " << cen+1 
	       << ": cluster " << cnum << endl;
      }
      cst = i;
      cen = i;
      cnum = expanded[i];
    } else {
      cen++;
    }
  }
  if (cst < cen) {		// must be an unfinished cluster
    cnum = expanded[nframes-1];	// last cluster in the file
    if (cnum < 0) {		// non-speech
      segout << "from " << cst+1 << " to " << cen+1 
	     << ": cluster <nonspeech>" << endl;
    } else {
      segout << "from " << cst+1 << " to " << cen+1 
	     << ": cluster " << cnum << endl;
    }
  }
  segout.close();
}

////////////////////////////////////////////////////////////
int HMM::nU(void) {
  int n_usable = 0;
  for (int cl = 0; cl < nClass; cl++) {
    if (U[cl])
      n_usable++;
  }
  return n_usable;
}

////////////////////////////////////////////////////////////
void
HMM::PrintHMM(FILE *fout) {
  int n_usable = 0;
  n_usable = nU();
  fprintf(fout, "%s%d\n", message1, n_usable); // "Number of clusters: "
  for(int cl = 0; cl < nClass; cl++) {
    if (U[cl]) {
      fprintf(fout, "%s%d\n", message2, cl); // "Cluster "
      gmm[cl].PrintGMM(fout);
    }
  }
}

////////////////////////////////////////////////////////////
void
HMM::ReadHMM(FILE *fin, const int nC, const int d) {
  nClass = nC; // number of clusters
  DIM = d; // dimension of feature vectors

  const int nBuffer = 30;
  char buffer[nBuffer];
  //char temp[nBuffer];

  int n;
  do {
    // "Number of cluster: "
    /*
    fgets(buffer,nBuffer,fin); // get line
    strncpy(temp,buffer,len1); temp[len1] = '\0'; // compare substring
    if (strcmp(temp,message1)) { // if match
      memcpy(temp,buffer[len1],nBuffer-len1); // convert rest of line to int
      n = atoi(temp); // number of clusters
    }
    */

    n = -1; // initialize to propagate do-while loop
    fgets(buffer,len1,fin); // get label
    if (strcmp(buffer,message1)) { // if match
      fgets(buffer,nBuffer,fin); // convert rest of line to int
      n = atoi(buffer); // number of clusters
    }
  } while ((n != nClass) || (!feof(fin))); // stop when the right number of clusters are found

  if (feof(fin)) {
    printf("Could not find HMM with %i clusters\n", nClass);
    exit(0);
  }

  // ignore old cluster numbers and rename to 0 .. (nClass - 1)
  delete []U; U = new int[nClass]; // create new usable clusters vector
  delete gmm; gmm = new GMM[nClass]; // create new GMM
  for (int a = 0; a < nClass; a++) { // for each cluster
    // "Cluster "
    /*
    fgets(buffer,nBuffer,fin);
    strncpy(temp,buffer,len2); temp[len2] = '\0'; // compare substring
    if (!strcmp(temp,message2)) { // if not a match
      printf("Tried to read \"%sx\" and instead found \"%s\"", message2, buffer);
      exit(0);
    }
    */

    fgets(buffer,len2,fin);
    if (!strcmp(buffer,message2)) { // if not a match
      printf("Tried to read \"%sx\" and instead found \"%s\"", message2, buffer);
      exit(0);
    }
    gmm[a].ReadGMM(fin, DIM); // read in GMM
    U[a] = 1; // set usable flag
  }
}
