#!/usr/bin/env python
# -*-Mode:Python-*-

import user, commands
import sys
import os
import getopt
import time

# Global vars
infile = basename = chunkfile = workbase = rttmprog = None
framesize = 0.01
offset = 0.0                            # time offset from input UEM (secs)
mindur = 2.0
ngauss = 5
niclus = 40
featdim = 12
segprog = './cluster'

def usage():
    print "\nUsage:",sys.argv[0],"<options>",
    print """
Where <options> are:
    -i|--infile <featfile>          -the input feature file
    -B|--basename <name>            -the base name of the file you are
                                     testing (e.g. 20010206_1830_1900_ABC_WNT)

  These are optional ([] = default values):
    -w|--workbase <dir>             -base path to working directory [expt]
    -S|--segprog <prog>             -segmentation program [./cluster]
    -R|--rttmprog <prog>            -program to convert from .seg to .rttm [None]
    -c|--chunkfile <file>           -name of the chunk info file [None]
    -f|--framesize <n>              -size of frame step in seconds [0.01]
    -d|--mindur <n>                 -min duration in seconds [2.0]
    -g|--ngauss <n>                 -number of gaussians per cluster [5]
    -n|--niclus <n>                 -number of initial clusters [40]
    -a|--featdim <n>                -dimensionality of feature vectors [12]
    -o|--offset <secs>              -time offset in secs, from input UEM [0.0]
"""
    sys.exit(2)

def parse_cmdline(argv):
    global infile, basename
    global segprog, rttmprog
    global framesize, mindur, ngauss, niclus, featdim, chunkfile, workbase
    global offset
    
    cmdoptions = ["infile=","basename=",
                  "segprog=", "rttmprog=",
                  "framesize=","mindur=","ngauss=","niclus=",
                  "featdim=","chunkfile=","workbase=","offset="]

    if len(argv) < 2:
        usage()
    try:
        opts,args = getopt.getopt(sys.argv[1:],
                                  "i:s:B:w:S:R:c:f:d:g:n:a:o:", cmdoptions)
    except getopt.GetoptError:
        print "Bad command line option"
        usage()

    for o,a in opts:
        if o in ("-i","--infile"):
            infile = a
            if not os.path.exists(infile):
                time.sleep(5)           # sleep and try again
                if not os.path.exists(infile):
                    print "%s: Error finding input feature file"%sys.argv[0]
                    sys.exit(1)
        if o in ("-B","--basename"):
            basename = a
        if o in ("-w","--workbase"):
            workbase = a
            if not os.path.isdir(workbase):
                print "Error finding dir:",workbase
                sys.exit(1)
            if(workbase[len(workbase)-1] != '/'):
                workbase = workbase + '/'
        if o in ("-S","--segprog"):
            segprog = a
        if o in ("-R","--rttmprog"):
            rttmprog = a
        if o in ("-c","--chunkfile"):
            chunkfile = a
            if not os.path.exists(chunkfile):
                print "Error finding chunk file"
                sys.exit(1)            
        if o in ("-f","--framesize"):
            framesize = float(a)
        if o in ("-d","--mindur"):
            mindur = float(a)
        if o in ("-g","--ngauss"):
            ngauss = int(a)
        if o in ("-n","--niclus"):
            niclus = int(a)
        if o in ("-a","--featdim"):
            featdim = int(a)
        if o in ("-o","--offset"):
            offset = float(a)
            if (offset < 0.0):
                print "Can't have negative offset."
                usage()

    if infile == None:
        print "You must specify an input feature file."
        usage()
    if workbase == None:
        workbase = 'expt/'
    if framesize <= 0.0:
        print "Illegal value for framesize. Must be > 0.0"
        usage()
    if mindur <= 0.0 or mindur < framesize:
        print "Illegal value for minimum duration. Must be > 0.0",
        print "and > framesize."
        usage()
    if ngauss <= 0:
        print "Illegal value for ngauss. Must be > 0"
        usage()
    if niclus <= 0:
        print "Illegal value for number of begining clusters. Must be > 0"
        usage()
    if featdim <= 0:
        print "Illegal value for feature vector dimension. Must be > 0"
        usage()
    if basename == None:
        print "No base name specified"
        usage()

    if not os.path.exists(segprog):
        print "Error finding segmentation prog:", segprog
        sys.exit(1)

    if rttmprog and not os.path.exists(rttmprog):
        print "Error finding seg to rttm prog:", rttmprog
        sys.exit(1)

def makepath(path):

    """ creates missing directories for the given path and
        returns a normalized absolute version of the path.

    - if the given path already exists in the filesystem
      the filesystem is not modified.

    - otherwise makepath creates directories along the given path
      using the dirname() of the path. You may append
      a '/' to the path if you want it to be a directory path.

    from holger@trillke.net 2002/03/18
    Added try statement to catch exceptions caused by weird timing.
       -wooters@icsi.berkeley.edu 2003/08/29
    """

    from os import makedirs
    from os.path import normpath,dirname,exists,abspath

    dpath = normpath(dirname(path))
    os.system("mkdir -p %s"%dpath)
    return normpath(abspath(path))

def docmd(cmd):
    result = commands.getstatusoutput(cmd)
    if result[0] != 0:
        sys.exit("Error running: %s:\n%s"%(cmd,result[1]))
    return result[1]

if __name__=='__main__':
    debug = 0
    
    parse_cmdline(sys.argv)

    # create the working directory
    workdir = os.path.join(workbase,basename)
    docmd("mkdir -p %s"%workdir)
    #workdir = makepath(workdir)+"/"         # make sure dir exists
    print "Workdir:",workdir
    sys.stdout.flush()
    
    # create the log file
    logfile = os.path.join(workdir,"log")
    logfh = open(logfile,'w')

    starttime = time.time()

    # log the user
    print >> logfh, "Process started by: %s" % os.environ['USER']
    # log the host name
    print >> logfh, "Running on: %s" % os.uname()[1]
    logfh.flush()
    
    # figure out how many frames of data to use for the minimum duration
    # need to make sure this is a multiple of 10
    # (JGF: Why does this have to be a multiple of 10?)
    nframes = int(mindur / framesize) + 5
    nframes = int(nframes / 10)
    nframes = nframes * 10

    ##################################################
    #
    # Do the segmentation
    #
    segout = os.path.join(workdir,"seg.out")
    if (chunkfile == None):
        segcmd = "%s %d %d %d %d %s %s > %s" % (segprog,niclus,ngauss,featdim,
                                                nframes,infile,segout,
                                                logfile+".segprog")
    else:
        segcmd = "%s %d %d %d %d %s %s %s > %s" % (segprog,niclus,
                                                   ngauss,featdim,
                                                   nframes,infile,
                                                   segout,chunkfile,
                                                   logfile+".segprog")

    print >> logfh, "\nRunning:\n%s" % segcmd
    print "Running:",segcmd
    
    rv = os.system(segcmd)
    if rv != 0:
        print "Error running",segprog
        print >> logfh, "Error running",segprog
        sys.exit(1)

    ##################################################
    #
    # convert the .seg output file to rttm format
    #
    if rttmprog:
        fps = 1.0/framesize
        outputrttmfile = os.path.join(workdir, basename+".rttm")
        convertcmd = "%s -f %s -s %s -o %f -b %s > %s" % (rttmprog,fps,
                                                          segout,offset, 
                                                          basename,
                                                          outputrttmfile)
        print "Running:",convertcmd
        print >> logfh, "\nRunning:\n%s" % convertcmd
        rv = os.system(convertcmd)
        if rv != 0:
            print "Error running ",rttmprog
            print >> logfh, "Error running ",rttmprog
            sys.exit(1)

    ##################################################
            
    endtime = time.time()
    runtime = "%.2f" % (endtime-starttime)
    xrt = "%.2f" % ((endtime-starttime)/1800.0)
    print >> logfh, "\nTotal Runtime (in seconds):",runtime
    print >> logfh, "   xRT (assuming 0.5 hour file):",xrt

    logfh.close()

    sys.exit(0)
    
